@@ -2,21 +2,25 @@ from multiprocessing import Process
 import threading
 import queue
 import time as ttime
+import os
 import asyncio
 from functools import partial
 import logging
 
 from .comms import PipeJsonRpcReceive
+from .output_streaming import setup_console_output_redirection
+
+from event_model import RunRouter
 
 import msgpack
 import msgpack_numpy as mpn
 
 from .profile_ops import (
-    load_profile_collection,
+    load_worker_startup_code,
     plans_from_nspace,
     devices_from_nspace,
     load_allowed_plans_and_devices,
-    parse_plan,
+    prepare_plan,
 )
 
 logger = logging.getLogger(__name__)
@@ -34,13 +38,16 @@ class RunEngineWorker(Process):
         `args` and `kwargs` of the `multiprocessing.Process`
     """
 
-    def __init__(self, *args, conn, config=None, **kwargs):
+    def __init__(self, *args, conn, config=None, msg_queue=None, log_level=logging.DEBUG, **kwargs):
 
         if not conn:
             raise RuntimeError("Invalid value of parameter 'conn': %S.", str(conn))
 
         super().__init__(*args, **kwargs)
 
+        self._log_level = log_level
+        self._msg_queue = msg_queue
+
         # The end of bidirectional Pipe assigned to the worker (for communication with Manager process)
         self._conn = conn
 
@@ -68,13 +75,20 @@ class RunEngineWorker(Process):
         self._re_report_lock = None  # threading.Lock
 
         # Class that supports communication over the pipe
-        self._comm_to_manager = PipeJsonRpcReceive(conn=self._conn, name="RE Watchdog-Manager Comm")
+        self._comm_to_manager = PipeJsonRpcReceive(conn=self._conn, name="RE Worker-Manager Comm")
 
         self._db = None
-        self._config = config or {}
+
+        # Note: 'self._config' is a private attribute of 'multiprocessing.Process'. Overriding
+        #   this variable may lead to unpredictable and hard to debug issues.
+        self._config_dict = config or {}
         self._allowed_plans, self._allowed_devices = {}, {}
 
-        self._re_namespace, self._existing_plans, self._existing_devices = {}, {}, {}
+        # The list of runs that were opened as part of execution of the currently running plan.
+        # Initialized with 'RunList()' in 'run()' method.
+        self._active_run_list = None
+
+        self._re_namespace, self._plans_in_nspace, self._devices_in_nspace = {}, {}, {}
 
     def _execute_plan(self, plan, is_resuming):
         """
@@ -111,6 +125,9 @@ class RunEngineWorker(Process):
                 # Include RE state
                 self._re_report["re_state"] = str(self._RE._state)
 
+                # Clear the list of active runs
+                self._active_run_list.clear()
+
         except BaseException as ex:
             with self._re_report_lock:
 
@@ -133,10 +150,13 @@ class RunEngineWorker(Process):
                     self._re_report["success"] = False
                     self._state["running_plan_completed"] = True
 
+                    # Clear the list of active runs (don't clean the list for the paused plan).
+                    self._active_run_list.clear()
+
                 # Include RE state
                 self._re_report["re_state"] = str(self._RE._state)
 
-        logger.debug("Finished execution of a task")
+        logger.debug("Finished execution of the task")
 
     def _load_new_plan(self, plan_info):
         """
@@ -160,32 +180,37 @@ class RunEngineWorker(Process):
         logger.info("Starting a plan '%s'.", plan_info["name"])
 
         try:
-            plan_parsed = parse_plan(
-                plan_info, allowed_plans=self._existing_plans, allowed_devices=self._existing_devices
+            plan_parsed = prepare_plan(
+                plan_info,
+                plans_in_nspace=self._plans_in_nspace,
+                devices_in_nspace=self._devices_in_nspace,
+                allowed_plans=self._allowed_plans,
+                allowed_devices=self._allowed_devices,
             )
 
-            plan_func = plan_parsed["name"]
+            plan_func = plan_parsed["callable"]
             plan_args_parsed = plan_parsed["args"]
             plan_kwargs_parsed = plan_parsed["kwargs"]
+            plan_meta_parsed = plan_parsed["meta"]
 
-            def get_plan(plan_func, plan_args, plan_kwargs):
+            def get_plan(plan_func, plan_args, plan_kwargs, plan_meta):
                 def plan():
                     if self._RE._state == "panicked":
                         raise RuntimeError(
-                            "Run Engine is in the 'panicked' state. "
-                            "You need to recreate the environment before you can run plans."
+                            "Run Engine is in the 'panicked' state. The environment must be "
+                            "closed and opened again before plans could be executed."
                         )
                     elif self._RE._state != "idle":
                         raise RuntimeError(
                             f"Run Engine is in '{self._RE._state}' state. Stop or finish any running plan."
                         )
                     else:
-                        result = self._RE(plan_func(*plan_args, **plan_kwargs))
+                        result = self._RE(plan_func(*plan_args, **plan_kwargs), **plan_meta)
                     return result
 
                 return plan
 
-            plan = get_plan(plan_func, plan_args_parsed, plan_kwargs_parsed)
+            plan = get_plan(plan_func, plan_args_parsed, plan_kwargs_parsed, plan_meta_parsed)
             # 'is_resuming' is true (we start a new plan that is supposedly runs to completion
             #   as opposed to aborting/stopping/halting a plan)
         except Exception as ex:
@@ -247,17 +272,19 @@ class RunEngineWorker(Process):
         """
         Returns the state information of RE Worker environment.
         """
-        plan_uid = self._state["running_plan"]["plan_uid"] if self._state["running_plan"] else None
+        item_uid = self._state["running_plan"]["item_uid"] if self._state["running_plan"] else None
         plan_completed = self._state["running_plan_completed"]
         re_state = str(self._RE._state) if self._RE else "null"
         env_state = self._state["environment_state"]
         re_report_available = self._re_report is not None
+        run_list_updated = self._active_run_list.is_changed()  # True - updates are available
         msg_out = {
-            "running_plan_uid": plan_uid,
+            "running_item_uid": item_uid,
             "running_plan_completed": plan_completed,
             "re_report_available": re_report_available,
             "re_state": re_state,
             "environment_state": env_state,
+            "run_list_updated": run_list_updated,
         }
         return msg_out
 
@@ -273,12 +300,22 @@ class RunEngineWorker(Process):
         self._re_report = None
         return msg_out
 
+    def _request_run_list_handler(self):
+        """
+        Returns the list of runs for the currently executed plan and clears the state
+        of the list. The list can be requested at any time, but it is recommended that
+        the state of the list is checked first (`re_report` field `run_list_updated`)
+        and update is loaded only if updates exist (`run_list_updated` is True).
+        """
+        msg_out = {"run_list": self._active_run_list.get_run_list(clear_state=True)}
+        return msg_out
+
     def _command_close_env_handler(self):
         """
         Close RE Worker environment in orderly way.
         """
         # Stop the loop in main thread
-        logger.info("Closing RE Worker environment")
+        logger.info("Closing RE Worker environment ...")
         # TODO: probably the criteria on when the environment could be more precise.
         #       For now simply assume that we can not close the environment in which
         #       Run Engine is running using this method. Different method that kills
@@ -323,8 +360,8 @@ class RunEngineWorker(Process):
         """
         Initiate execution of a new plan.
         """
-        logger.info("Starting execution of a plan")
-        # TODO: refine the criteria of acceptance of the new plan.
+        logger.info("Starting execution of a plan ...")
+        # TODO: refine the conditions that are verified before a new plan is accepted.
         invalid_state = 0
         if not self._execution_queue.empty():
             invalid_state = 1
@@ -367,7 +404,7 @@ class RunEngineWorker(Process):
         Pause running plan. Options: `deferred` and `immediate`.
         """
         # Stop the loop in main thread
-        logger.info("Pausing Run Engine")
+        logger.info("Pausing Run Engine ...")
         pausing_options = ("deferred", "immediate")
         # TODO: the question is whether it is possible or should be allowed to pause a plan in
         #       any other state than 'running'???
@@ -424,7 +461,7 @@ class RunEngineWorker(Process):
             status = "accepted"
         else:
             status = "rejected"
-            err_msg = "Run Engine must be in 'idle' state to continue. " f"The state is '{self._RE._state}'"
+            err_msg = f"Run Engine must be in 'idle' state to continue. The state is '{self._RE._state}'"
 
         msg_out = {"status": status, "err_msg": err_msg}
         return msg_out
@@ -457,10 +494,26 @@ class RunEngineWorker(Process):
         Overrides the `run()` function of the `multiprocessing.Process` class. Called
         by the `start` method.
         """
+        setup_console_output_redirection(msg_queue=self._msg_queue)
+
+        logging.basicConfig(level=max(logging.WARNING, self._log_level))
+        logging.getLogger(__name__).setLevel(self._log_level)
+
         success = True
 
+        from .profile_tools import set_re_worker_active, clear_re_worker_active
+
+        # Set the environment variable indicating that RE Worker is active. Status may be
+        #   checked using 'is_re_worker_active()' in startup scripts or modules.
+        set_re_worker_active()
+
+        from .plan_monitoring import RunList, CallbackRegisterRun
+
+        self._active_run_list = RunList()  # Initialization should be done before communication is enabled.
+
         self._comm_to_manager.add_method(self._request_state_handler, "request_state")
         self._comm_to_manager.add_method(self._request_plan_report_handler, "request_plan_report")
+        self._comm_to_manager.add_method(self._request_run_list_handler, "request_run_list")
         self._comm_to_manager.add_method(self._command_close_env_handler, "command_close_env")
         self._comm_to_manager.add_method(self._command_confirm_exit_handler, "command_confirm_exit")
         self._comm_to_manager.add_method(self._command_run_plan_handler, "command_run_plan")
@@ -477,11 +530,9 @@ class RunEngineWorker(Process):
         from bluesky.run_engine import get_bluesky_event_loop
         from bluesky.callbacks.best_effort import BestEffortCallback
         from bluesky_kafka import Publisher as kafkaPublisher
-        from databroker import Broker
+        from bluesky.utils import PersistentDict
 
-        # TODO: subscribe to local databroker (currently there is no way to access
-        #    'temp' databroker from outside the process). Production version will use Kafka.
-        self._db = Broker.named("temp")
+        from .profile_tools import global_user_namespace
 
         # TODO: TC - Do you think that the following code may be included in RE.__init__()
         #   (for Python 3.8 and above)
@@ -489,33 +540,38 @@ class RunEngineWorker(Process):
         loop = get_bluesky_event_loop()
         asyncio.set_event_loop(loop)
 
-        def init_namespace():
-            self._re_namespace = {}
-            self._existing_plans = {}
-            self._existing_devices = {}
+        try:
+            keep_re = self._config_dict["keep_re"]
+            startup_dir = self._config_dict.get("startup_dir", None)
+            startup_module_name = self._config_dict.get("startup_module_name", None)
+            startup_script_path = self._config_dict.get("startup_script_path", None)
 
-        if "profile_collection_path" not in self._config:
-            logger.warning("Path to profile collection was not specified. No profile collection will be loaded.")
-            init_namespace()
-        else:
-            path = self._config["profile_collection_path"]
-            logger.info("Loading beamline profile collection from directory '%s' ...", path)
-            try:
-                self._re_namespace = load_profile_collection(path)
-                self._existing_plans = plans_from_nspace(self._re_namespace)
-                self._existing_devices = devices_from_nspace(self._re_namespace)
-                logger.info("Beamline profile collection was loaded completed.")
-            except Exception as ex:
-                logger.exception(
-                    "Failed to start RE Worker environment. Error while " "loading profile collection: %s.",
-                    str(ex),
+            self._re_namespace = load_worker_startup_code(
+                startup_dir=startup_dir,
+                startup_module_name=startup_module_name,
+                startup_script_path=startup_script_path,
+                keep_re=keep_re,
+            )
+
+            if keep_re and ("RE" not in self._re_namespace):
+                raise RuntimeError(
+                    "Run Engine is not created in the startup code and 'keep_re' option is activated."
                 )
-                success = False
+            self._plans_in_nspace = plans_from_nspace(self._re_namespace)
+            self._devices_in_nspace = devices_from_nspace(self._re_namespace)
+            logger.info("Startup code loading was completed")
+
+        except Exception as ex:
+            logger.exception(
+                "Failed to start RE Worker environment. Error while loading startup code: %s.",
+                str(ex),
+            )
+            success = False
 
         # Load lists of allowed plans and devices
         logger.info("Loading the lists of allowed plans and devices ...")
-        path_pd = self._config["existing_plans_and_devices_path"]
-        path_ug = self._config["user_group_permissions_path"]
+        path_pd = self._config_dict["existing_plans_and_devices_path"]
+        path_ug = self._config_dict["user_group_permissions_path"]
         try:
             self._allowed_plans, self._allowed_devices = load_allowed_plans_and_devices(
                 path_existing_plans_and_devices=path_pd, path_user_group_permissions=path_ug
@@ -529,22 +585,63 @@ class RunEngineWorker(Process):
             logger.info("Instantiating and configuring Run Engine ...")
 
             try:
-                self._RE = RunEngine({})
+                # Make RE namespace available to the plan code.
+                global_user_namespace.set_user_namespace(user_ns=self._re_namespace, use_ipython=False)
 
-                bec = BestEffortCallback()
-                self._RE.subscribe(bec)
+                if self._config_dict["keep_re"]:
+                    # Copy references from the namespace
+                    self._RE = self._re_namespace["RE"]
+                    self._db = self._re_namespace.get("RE", None)
+                else:
+                    # Instantiate a new Run Engine and Data Broker (if needed)
+                    md = {}
+                    if self._config_dict["use_persistent_metadata"]:
+                        # This code is temporarily copied from 'nslsii' before better solution for keeping
+                        #   continuous sequence Run ID is found. TODO: continuous sequence of Run IDs.
+                        directory = os.path.expanduser("~/.config/bluesky/md")
+                        os.makedirs(directory, exist_ok=True)
+                        md = PersistentDict(directory)
 
-                self._RE.subscribe(self._db.insert)
+                    self._RE = RunEngine(md)
+                    self._re_namespace["RE"] = self._RE
 
-                if "kafka" in self._config:
+                    def factory(name, doc):
+                        # Documents from each run are routed to an independent
+                        #   instance of BestEffortCallback
+                        bec = BestEffortCallback()
+                        return [bec], []
+
+                    # Subscribe to Best Effort Callback in the way that works with multi-run plans.
+                    rr = RunRouter([factory])
+                    self._RE.subscribe(rr)
+
+                    # Subscribe RE to databroker if config file name is provided
+                    self._db = None
+                    if "databroker" in self._config_dict:
+                        config_name = self._config_dict["databroker"].get("config", None)
+                        if config_name:
+                            logger.info("Subscribing RE to Data Broker using configuration '%s'.", config_name)
+                            from databroker import Broker
+
+                            self._db = Broker.named(config_name)
+                            self._re_namespace["db"] = self._db
+
+                            self._RE.subscribe(self._db.insert)
+
+                # Subscribe Run Engine to 'CallbackRegisterRun'. This callback is used internally
+                #   by the worker process to keep track of the runs that are open and closed.
+                run_reg_cb = CallbackRegisterRun(run_list=self._active_run_list)
+                self._RE.subscribe(run_reg_cb)
+
+                if "kafka" in self._config_dict:
                     logger.info(
                         "Subscribing to Kafka: topic '%s', servers '%s'",
-                        self._config["kafka"]["topic"],
-                        self._config["kafka"]["bootstrap"],
+                        self._config_dict["kafka"]["topic"],
+                        self._config_dict["kafka"]["bootstrap"],
                     )
                     kafka_publisher = kafkaPublisher(
-                        topic=self._config["kafka"]["topic"],
-                        bootstrap_servers=self._config["kafka"]["bootstrap"],
+                        topic=self._config_dict["kafka"]["topic"],
+                        bootstrap_servers=self._config_dict["kafka"]["bootstrap"],
                         key="kafka-unit-test-key",
                         # work with a single broker
                         producer_config={"acks": 1, "enable.idempotence": False, "request.timeout.ms": 5000},
@@ -552,6 +649,12 @@ class RunEngineWorker(Process):
                     )
                     self._RE.subscribe(kafka_publisher)
 
+                if "zmq_data_proxy_addr" in self._config_dict:
+                    from bluesky.callbacks.zmq import Publisher
+
+                    publisher = Publisher(self._config_dict["zmq_data_proxy_addr"])
+                    self._RE.subscribe(publisher)
+
                 self._execution_queue = queue.Queue()
 
                 self._state["environment_state"] = "ready"
@@ -561,7 +664,7 @@ class RunEngineWorker(Process):
                 logger.exception("Error occurred while initializing the environment: %s.", str(ex))
 
         if success:
-            logger.info("RE Environment is ready.")
+            logger.info("RE Environment is ready")
             self._execute_in_main_thread()
         else:
             self._exit_event.set()
@@ -573,8 +676,12 @@ class RunEngineWorker(Process):
         while not self._exit_confirmed_event.is_set():
             ttime.sleep(0.02)
 
+        # Clear the environment variable indicating that RE Worker is active. It is an optional step
+        #   since the process is about to close, but we still do it for consistency.
+        clear_re_worker_active()
+
         self._RE = None
 
         self._comm_to_manager.stop()
 
-        logger.info("Run Engine environment was closed successfully.")
+        logger.info("Run Engine environment was closed successfully")
